/***************************************************************************
 *
 * $Id: wtmpmond.c,v 1.2 2004/09/14 18:19:00 emalkowski Exp $
 *
 * Desc:      Daemon to monitor login and logout events on a system and
 *            report the events to syslog.  This is accomplished by
 *            polling the /var/log/wtmp file and matching login and logout
 *            records as they occur.  This daemon is intended to be used
 *            along with a SNARE and syslog configuration to meet the
 *            NISPOM requirements for the Debian linux computers in
 *            a closed (classified) lab area.
 *
 * Author:    Eric Malkowski
 * E-mail:    malk@gweep.net
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <utmp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <syslog.h>

#define WTMPMOND_VERSION_STR	"0.0.2"

/* Default full path to wtmp file */

#define DEFAULT_WTMP_FILE	"/var/log/wtmp"

/* Complain about being unable to open /var/log/wtmp once every hour */

#define ONE_HOUR_SECONDS	(60 * 60)

/* Poll /var/log/wtmp fd every 5 seconds for login / logout events */

#define WTMP_POLL_TIME		5

/* When unable to open /var/log/wtmp, retry every 60 seconds */

#define WTMP_RETRY_OPEN_TIME	60

#ifndef TRUE
#define TRUE		1
#endif

#ifndef FALSE
#define FALSE		0
#endif

#define MAX_UT_TYPE		10

char *ut_type_names[] = { "UT_UNKNOWN",
                          "RUN_LVL",
                          "BOOT_TIME",
                          "NEW_TIME",
                          "OLD_TIME",
                          "INIT_PROCESS",
                          "LOGIN_PROCESS",
                          "USER_PROCESS",
                          "DEAD_PROCESS",
                          "ACCOUNTING",
                          NULL };

struct ut_list_entry
{
   struct utmp            ut;
   struct ut_list_entry  *next;
} *ut_list_ptr = NULL;

int on_console = FALSE;

/****************************************************************************/

/* When we've read a piece of a utmp record and then read 0 bytes (EOF), we
   try up to 3 more times w/ a one second pause in between in hopes the rest
   of the record will come in.  This should never happen since the wtmp
   file is locked for reading and writing w/ standard file locks, but this
   is defensive code in case some wtmp user doesn't grab the lock */

#define PARTIAL_READ_RETRY_TIME		1
#define PARTIAL_READ_MAX_RETRIES	3

int read_bytes(int fd, void *void_ptr, int amount)
{
   char *buf_ptr = (char *) void_ptr;
   int bytes_remaining = amount;
   int total_bytes = 0;
   int partial = 0;
   int bytes_read;

   while(TRUE)
   {
      bytes_read = read(fd, buf_ptr + total_bytes, bytes_remaining);

      if (bytes_read < 0)
      {
         syslog(LOG_ERR, "read() failed, errno: %d (%s)",
                errno, strerror(errno));
         return(-1);
      }

      total_bytes += bytes_read;
      bytes_remaining -= bytes_read;
 
      if (bytes_read == 0)
      {
         if ((total_bytes > 0) && (bytes_remaining > 0))
         {
            partial++;
       
            if (partial == PARTIAL_READ_MAX_RETRIES)
               break;

            sleep(PARTIAL_READ_RETRY_TIME);
            continue;
         }
         else
            break;
      }
      else
      {
         partial = 0;
      }

      if (total_bytes == amount)
         break;
   }

   if (total_bytes == 0)
      return(0);

   if (total_bytes != amount)
      return(-1);

   return(total_bytes);
}

int lock_file(int fd)
{
   struct flock fl;

   memset(&fl, 0, sizeof(struct flock));
   fl.l_type = F_RDLCK; 
   fl.l_whence = SEEK_SET;

   if (fcntl(fd, F_SETLKW, &fl) < 0)
   {
      syslog(LOG_ERR, "fcntl(F_SETLKW) for lock failed, errno: %d (%s)",
             errno, strerror(errno));
      syslog(LOG_ERR, "exiting due to fatal error");
      exit(1);
   }
  
   return(0);
}

int unlock_file(int fd)
{
   struct flock fl;

   memset(&fl, 0, sizeof(struct flock));
   fl.l_type = F_UNLCK;
   fl.l_whence = SEEK_SET;

   if(fcntl(fd, F_SETLKW, &fl) < 0)
   {
      syslog(LOG_ERR, "fcntl(F_SETLKW) for unlock failed, errno: %d (%s)",
             errno, strerror(errno));
      syslog(LOG_ERR, "exiting due to fatal error");
      exit(1);
   }

   return(0);
}

int file_changed(int fd, char *filename)
{
   struct stat fdstat;
   struct stat fnstat;

   if (fstat(fd, &fdstat) < 0)
   { 
      syslog(LOG_ERR, "fstat() of fd: %d failed, errno: %d (%s)",
             fd, errno, strerror(errno));

      /* Indicate a change when we have problems w/ fstat 
         so caller can close and re-open */

      return(TRUE);
   }

   if (stat(filename, &fnstat) < 0)
   {
      /* Indicate a change */
      return(TRUE);
   }

   if (fdstat.st_ino != fnstat.st_ino)
      return(TRUE);
 
   if (fdstat.st_dev != fnstat.st_dev)
      return(TRUE);

   /* Inodes and devices still match, fd didn't change */

   return(FALSE);
}

void arm_signal(int signum);

void generic_signal_handler(int signum)
{
   switch(signum)
   {
      case SIGHUP:
      case SIGTERM:
      case SIGQUIT:
         syslog(LOG_NOTICE, "exiting.");
         exit(0);
      break;

      default:
         syslog(LOG_WARNING,
                "generic signal handler: caught signal %d",
                 signum);
      break;
   }

   arm_signal(signum);
}

void arm_signal(int signum)
{
   struct sigaction action;

   memset(&action, 0, sizeof(struct sigaction));
   action.sa_handler = generic_signal_handler;

   sigaction(signum, &action, NULL);
}

void setup_signals()
{
   /* When in console mode, we want CTRL-C to work, otherwise
      catch SIGINT and don't die */

   if (!on_console)
      arm_signal(SIGINT);

   arm_signal(SIGTERM);
   arm_signal(SIGQUIT);
   arm_signal(SIGURG);
   arm_signal(SIGPIPE);
   arm_signal(SIGUSR1);

}

void daemonize()
{
   pid_t pid;

   if ((pid = fork()) == -1)
   {
      syslog(LOG_ERR,
              "fork() failed, errno: %d (%s), exiting...",
              errno, strerror(errno));
      exit(1);
   }

   if (pid != 0)
   {
      /* Parent exits clean */
      exit(0);
   }

   /* Child processing -- disconnect from controlling terminal */

   close(STDIN_FILENO);
   close(STDOUT_FILENO);
   close(STDERR_FILENO);

   chdir("/");

   setsid();

   /* We are now a daemon */
}

/******************************************************************************/

#define TIME_STR_LEN		64

int main(int argc, char *argv[])
{
   int fd = -1;
   int initialized = FALSE;
   int bytes_read;
   struct utmp utmp_entry;
   struct ut_list_entry *entry_ptr;
   struct ut_list_entry *cur_entry_ptr;
   struct ut_list_entry *prev_entry_ptr;
   struct in_addr *in_addr_ptr;
   time_t no_wtmp_stamp = 0;
   time_t cur_time;
   struct tm tm, *tm_ptr;
   int c;
   char *wtmp_filename = DEFAULT_WTMP_FILE;
   int online_time, online_time_days, online_time_hours,
       online_time_minutes, online_time_seconds;
   char id_str[5] = { 0, 0, 0, 0, 0};
   char log_msg[512];
   char time_str[TIME_STR_LEN];

   while((c = getopt(argc, argv, "cf:")) != -1)
   {
      switch(c)
      {
         case 'c':
            on_console = TRUE;
         break;

         case 'f':
            wtmp_filename = optarg;
         break;

         default:
            fprintf(stderr, "\nUsage: wtmpmond [ -c ] [ -f wtmpfile ]\n\n");
            exit(1);
         break;
      }
   }

   setup_signals();

   openlog("wtmpmond", LOG_NDELAY | LOG_PID, LOG_AUTH);

   if (!on_console)
      daemonize();

   syslog(LOG_NOTICE, "version %s started.", WTMPMOND_VERSION_STR);

   while(TRUE)
   {
      if (fd < 0)
      {
         fd = open(wtmp_filename, O_RDONLY);
 
         if (fd < 0)
         {
            time(&cur_time);

            if ((cur_time - no_wtmp_stamp) > ONE_HOUR_SECONDS)
            {
               syslog(LOG_WARNING, "Unable to open %s", wtmp_filename);
               no_wtmp_stamp = cur_time;
            }

            sleep(WTMP_RETRY_OPEN_TIME);
            continue;
         }
         else
         {
            no_wtmp_stamp = 0;
         }

         /* When we get here, we got the file opened */

         if (!initialized)
         {
            /* First time running, or re-syncing after a serious error
               dealing with /var/log/wtmp, skip over all old stuff to the
               end of the file.  */

            lock_file(fd);

            if (lseek(fd, 0, SEEK_END) < 0)
            {
               syslog(LOG_ERR, "lseek(SEEK_END) failed, errno: %d (%s)",
                      errno, strerror(errno));

               unlock_file(fd);

               close(fd);
               fd = -1;

               continue;
            }

            unlock_file(fd);

            initialized = TRUE;
         }
      }
      
      lock_file(fd);

      bytes_read = read_bytes(fd, &utmp_entry, sizeof(struct utmp));

      unlock_file(fd);

      if (bytes_read == 0)
      {
         if (file_changed(fd, wtmp_filename))
         {
            syslog(LOG_NOTICE, "INFO: %s rotated", wtmp_filename);
            close(fd);
            fd = -1;

            continue;
         }     

         if (on_console)
         {
            printf("At EOF - continue to poll\n");
         }

         sleep(WTMP_POLL_TIME);
         continue;
      }

      if (bytes_read < 0)
      {
         close(fd);
         fd = -1;

         initialized = FALSE;
 
         continue;
      }

      if (bytes_read != sizeof(struct utmp))
      { 
         syslog(LOG_ERR, "expected %d bytes, got %d, will attempt to re-sync", 
                 sizeof(struct utmp), bytes_read);
 
         close(fd);
         fd = -1;

         initialized = FALSE;
         continue;
      }

      /* Check if it's a time/date change entry by the date command */

      if ( ( ( (strncmp(utmp_entry.ut_line, "|", UT_LINESIZE) == 0) ||
               (strncmp(utmp_entry.ut_line, "{", UT_LINESIZE) == 0) ) ) &&
           (utmp_entry.ut_id[0] == 0) &&
           (strncmp(utmp_entry.ut_user, "date", UT_NAMESIZE) == 0) )
      {
         /* Not interested in date change entries
            see "man utmp" page for explanation of | and { chars
            with a ut_user called "date" */

         continue;
      }

      if (on_console)
      {
         if (utmp_entry.ut_type < MAX_UT_TYPE)
         {
            printf("Type: %s (%d)\n", 
                   ut_type_names[utmp_entry.ut_type], utmp_entry.ut_type);
         }
         else
         {
            printf("Type:   UNKNOWN (%d)\n", utmp_entry.ut_type);
         }


         printf ("Pid:   %d\n", utmp_entry.ut_pid);
         printf ("Line:  %s\n", utmp_entry.ut_line);

         memcpy(id_str, utmp_entry.ut_id, 4);

         printf ("Id:    %s\n", id_str);
         printf ("User:  %s\n", utmp_entry.ut_user);
         printf ("Host:  %s\n", utmp_entry.ut_host);
         printf ("Exit:  %d %d\n", utmp_entry.ut_exit.e_termination, utmp_entry.ut_exit.e_exit);
         printf ("Time:  %s", ctime(&(utmp_entry.ut_tv.tv_sec)));
         printf ("Sess:  0x%08x\n", (unsigned int) utmp_entry.ut_session);

         if (strncmp("pts", utmp_entry.ut_line, 3) == 0)
         {
            in_addr_ptr = (struct in_addr *) (utmp_entry.ut_addr_v6);
            printf ("IP:    %s\n\n", inet_ntoa(*in_addr_ptr));
         }
         else
         {
            printf ("IP:    \n\n");
         }
      }

      if (utmp_entry.ut_type == USER_PROCESS)
      {
         if (utmp_entry.ut_pid == 0)
            continue;

         if (utmp_entry.ut_line[0] == 0)
            continue;

         if (utmp_entry.ut_user[0] == 0)
            continue;

         strncpy(log_msg, utmp_entry.ut_user, UT_NAMESIZE);
 
         strcat(log_msg, " login on ");
         strncat(log_msg, utmp_entry.ut_line, UT_LINESIZE);

         if ( ((strncmp(utmp_entry.ut_line, "pts", 3) == 0) || 
               (strncmp(utmp_entry.ut_line, "ftp", 3) == 0)) &&
              (utmp_entry.ut_host[0] != 0))
         {
            strcat(log_msg, " from ");
            strncat(log_msg, utmp_entry.ut_host, UT_HOSTSIZE);
         }

         tm_ptr = localtime(&(utmp_entry.ut_tv.tv_sec));
         tm = *tm_ptr;

         strftime(time_str, TIME_STR_LEN, "%d-%b-%Y, %H:%M:%S %Z", &tm);

         strcat(log_msg, " at ");
         strcat(log_msg, time_str);

         syslog(LOG_NOTICE, log_msg);

         entry_ptr = 
            (struct ut_list_entry *) malloc(sizeof(struct ut_list_entry));
  
         if (entry_ptr == NULL)
         {
            syslog(LOG_ERR, "malloc() failed, skipping login record");
            continue;
         }

         memcpy(&(entry_ptr->ut), &utmp_entry, sizeof(struct utmp));

         entry_ptr->next = ut_list_ptr;
         ut_list_ptr = entry_ptr;
      }

      if (utmp_entry.ut_type == DEAD_PROCESS)
      {
         cur_entry_ptr = ut_list_ptr;
         prev_entry_ptr = NULL;

         while(cur_entry_ptr)
         {
            if (strncmp(cur_entry_ptr->ut.ut_line,
                        utmp_entry.ut_line, UT_LINESIZE) == 0)
            {

               strncpy(log_msg, cur_entry_ptr->ut.ut_user, UT_NAMESIZE);

               strcat(log_msg, " logged out from ");
               strncat(log_msg, utmp_entry.ut_line, UT_LINESIZE);

               tm_ptr = localtime(&(utmp_entry.ut_tv.tv_sec));
               tm = *tm_ptr;

               strftime(time_str, TIME_STR_LEN, "%d-%b-%Y, %H:%M:%S %Z", &tm);

               strcat(log_msg, " at ");
               strcat(log_msg, time_str);

               online_time = utmp_entry.ut_tv.tv_sec -
                             cur_entry_ptr->ut.ut_tv.tv_sec;

               online_time_days = online_time / 86400;
               online_time = online_time % 86400;

               online_time_hours = online_time / 3600;
               online_time = online_time % 3600;
  
               online_time_minutes = online_time / 60;
               online_time_seconds = online_time % 60;

               sprintf(time_str, " (on %d days, %02d:%02d:%02d)",
                       online_time_days, online_time_hours,
                       online_time_minutes, online_time_seconds);

               strcat(log_msg, time_str);
             
               syslog(LOG_NOTICE, log_msg);


               if (prev_entry_ptr)
                  prev_entry_ptr->next = cur_entry_ptr->next;
               else
                  ut_list_ptr = cur_entry_ptr->next;

               free(cur_entry_ptr);
 
               break;
            }

            prev_entry_ptr = cur_entry_ptr;
            cur_entry_ptr = cur_entry_ptr->next;
         }
      }
   }
}

/**************************************************************************
 *
 * $Log: wtmpmond.c,v $
 * Revision 1.2  2004/09/14 18:19:00  emalkowski
 * Add -f support and clean up some error messages / program exit paths
 *
 * Revision 1.1  2004/07/19 18:22:03  emalkowski
 * Initial revision
 *
 **************************************************************************/
